#ifndef DAEMON_H
#define DAEMON_H

#include <boost/thread/detail/thread.hpp>

#include "proxy.h"
#include "api-driver/structs.h"

namespace api_driver {
    class TerminalApiDaemon {
        std::mutex cpApiMutex;
        boost::thread daemon;
        proxy::CpProxy cp;

        void connectToApi();

        void run();

        std::shared_mutex stateMutex;
        obj::TerminalState state;
        obj::TerminalDeviceState stateDev;

        std::shared_mutex settingsMutex;
        obj::TerminalRxTxSettings settingsRxTx;
#if API_OBJECT_NETWORK_SETTINGS_ENABLE
        obj::TerminalNetworkSettings settingsNetwork;
#endif
#if API_OBJECT_QOS_SETTINGS_ENABLE
        obj::TerminalQosSettings settingsQos;
#endif

        std::shared_mutex firmwareMutex;
        obj::TerminalFirmwareVersion firmware;

    public:
#if API_OBJECT_STATISTICS_ENABLE
        obj::StatisticsLogger statsLogs;
#endif

        explicit TerminalApiDaemon();

        std::string getDeviceInitState();

        /**
         * Получение статистики, копирует текущие значения в структуры, переданные по указателю. Если передан пустой указатель, копирования не произойдет.
         */
        void getState(obj::TerminalState& dest);
        void getDeviceState(obj::TerminalDeviceState& dest);

        // /**
        //  * Получение настроек, копирует текущие значения в структуры, переданные по указателю. Если передан пустой указатель, копирования не произойдет.
        //  */
        void getSettingsRxTx(obj::TerminalRxTxSettings& dest);

        void getNetworkSettings(obj::TerminalNetworkSettings& dest);

        void getQosSettings(bool& isEnabled, std::string& json);

        void setSettingsRxTx(obj::TerminalRxTxSettings& s, bool readback = true);

        void setSettingsDpdi(DPDI_parmeters& s, bool readback = true);

        void setSettingsBucLnb(buc_lnb_settings& bucLnb, bool readback = true);

        void setQosSettings(bool enabled, const std::string& str, bool readback = true);

        void setNetworkSettings(obj::TerminalNetworkSettings& s, bool readback = true);

        obj::TerminalFirmwareVersion getFirmware();

        void resetPacketStatistics();

        void resetDefaultSettings();

        ~TerminalApiDaemon();
    };
}


#endif //DAEMON_H
